use std::{
    marker::PhantomData,
    ops::{Add, AddAssign, Sub, SubAssign},
};

use bevy::{
    ecs::component::Component,
    math::{IVec3, UVec3},
};

use super::Pos;

pub struct Region<T> {
    min: IVec3,
    max: IVec3,
    _marker: PhantomData<T>,
}

impl<T> Region<T> {
    pub fn new_unchecked(min: IVec3, max: IVec3) -> Self {
        Self {
            min,
            max,
            _marker: PhantomData,
        }
    }

    pub fn new_checked(min: Pos<T>, max: Pos<T>) -> Result<Self, ()> {
        let min: IVec3 = min.into();
        let max: IVec3 = max.into();
        if min.x <= max.x && min.y <= max.y && min.z <= max.z {
            Ok(Self::new_unchecked(min, max))
        } else {
            Err(())
        }
    }

    pub fn new(a: Pos<T>, b: Pos<T>) -> Self {
        let a: IVec3 = a.into();
        let b: IVec3 = b.into();
        let min = IVec3::new(a.x.min(b.x), a.y.min(b.y), a.z.min(b.z));
        let max = IVec3::new(a.x.max(b.x), a.y.max(b.y), a.z.max(b.z));
        Self::new_unchecked(min, max)
    }

    pub fn size(&self) -> UVec3 {
        (self.max - self.min + IVec3::ONE).as_uvec3()
    }

    pub fn contains(&self, pos: Pos<T>) -> bool {
        let pos: IVec3 = pos.into();
        (pos.x >= self.min.x && pos.x <= self.max.x)
            && (pos.y >= self.min.y && pos.y <= self.max.y)
            && (pos.z >= self.min.z && pos.z <= self.max.z)
    }

    pub fn intersects(&self, other: Self) -> bool {
        (other.max.x > self.min.x && other.min.x < self.max.x)
            && (other.max.y > self.min.y && other.min.y < self.max.y)
            && (other.max.z > self.min.z && other.min.z < self.max.z)
    }

    pub fn distance_to(&self, pos: Pos<T>) -> f32 {
        (self.distance_to_squared(pos) as f32).sqrt()
    }

    pub fn distance_to_squared(&self, pos: Pos<T>) -> i32 {
        let pos: IVec3 = pos.into();
        let clamped = pos.max(self.min).min(self.max);
        pos.distance_squared(clamped)
    }

    pub fn expand(&self, amount: i32) -> Result<Self, ()> {
        Self::new_checked(
            (self.min - IVec3::splat(amount)).into(),
            (self.max + IVec3::splat(amount)).into(),
        )
    }
}

impl<T> std::fmt::Display for Region<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{} to {}", self.min, self.max)
    }
}

impl<T> std::fmt::Debug for Region<T> {
    fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        fmt.debug_struct(stringify!(Region<T>))
            .field("min", &self.min)
            .field("max", &self.max)
            .finish()
    }
}

// Trait implementations that SHOULD be handled by #[derive]

use bevy::ecs::component::StorageType;
impl<T: Send + Sync + 'static> Component for Region<T> {
    const STORAGE_TYPE: StorageType = StorageType::Table;
}

impl<T> Default for Region<T> {
    fn default() -> Self {
        Self {
            min: IVec3::ZERO,
            max: IVec3::ZERO,
            _marker: PhantomData,
        }
    }
}

impl<T> Clone for Region<T> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<T> Copy for Region<T> {}

impl<T> PartialEq for Region<T> {
    fn eq(&self, other: &Self) -> bool {
        (self.min, self.max).eq(&(other.min, other.max))
    }
}

impl<T> Eq for Region<T> {}

impl<T> std::hash::Hash for Region<T> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        (self.min, self.max).hash(state)
    }
}

// Conversion

impl<T> From<Region<T>> for (Pos<T>, Pos<T>) {
    fn from(Region { min, max, .. }: Region<T>) -> Self {
        (min.into(), max.into())
    }
}

// Offsetting

impl<T> Add<IVec3> for Region<T> {
    type Output = Region<T>;
    fn add(self, rhs: IVec3) -> Self::Output {
        Self::new_unchecked(self.min + rhs, self.max + rhs)
    }
}
impl<T> Add<IVec3> for &Region<T> {
    type Output = Region<T>;
    fn add(self, rhs: IVec3) -> Self::Output {
        Region::new_unchecked(self.min + rhs, self.max + rhs)
    }
}

impl<T> Sub<IVec3> for Region<T> {
    type Output = Region<T>;
    fn sub(self, rhs: IVec3) -> Self::Output {
        Self::new_unchecked(self.min - rhs, self.max - rhs)
    }
}
impl<T> Sub<IVec3> for &Region<T> {
    type Output = Region<T>;
    fn sub(self, rhs: IVec3) -> Self::Output {
        Region::new_unchecked(self.min - rhs, self.max - rhs)
    }
}

impl<T> AddAssign<IVec3> for Region<T> {
    fn add_assign(&mut self, rhs: IVec3) {
        self.min += rhs;
        self.max += rhs;
    }
}

impl<T> SubAssign<IVec3> for Region<T> {
    fn sub_assign(&mut self, rhs: IVec3) {
        self.min -= rhs;
        self.max -= rhs;
    }
}
