use std::ops;

use bevy::math::IVec3;
use overload::overload;

use super::ChunkPos;

/// Helper function that creates a [`ZOrderIndex`], panicking if the supplied values
/// are not between [`ZOrderIndex::ELEMENT_MIN`] and [`ZOrderIndex::ELEMENT_MAX`].
pub fn zorder(x: i32, y: i32, z: i32) -> ZOrderIndex {
    ZOrderIndex::new(x, y, z).unwrap()
}

/// Encodes 3 signed 21-bit integers into a 64-bit integer, their bits interleaved.
///
/// This struct wraps an integer which represents an index into a space-filling curve called
/// [Z-Order Curve](https://en.wikipedia.org/wiki/Z-order_curve). This is also referred to as
/// Morton order, code, or encoding.
///
/// By interleaving the 3 sub-elements into a single integer, some amount of packing can be
/// achieved, at the loss of some bits per elements. For example, with this 64-bit integer,
/// 21 bits per elements are available (`2_097_152` distinct values), which may be enough to
/// represent block coordinates in a bloxel game world.
///
/// One upside of encoding separate coordinates into a single Z-Order index is that it can then
/// be effectively used to index into octrees, and certain operations such as bitwise shifting
/// are quite useful.
#[derive(Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct ZOrderIndex(u64);

impl ZOrderIndex {
    pub const ZERO: Self = Self::from_raw(0);

    const BITS_PER_ELEMENT: usize = (size_of::<u64>() * 8) / 3;
    pub const ELEMENT_MIN: i32 = !0 << (Self::BITS_PER_ELEMENT - 1);
    pub const ELEMENT_MAX: i32 = !Self::ELEMENT_MIN;

    const TOTAL_USABLE_BITS: usize = Self::BITS_PER_ELEMENT * 3;
    const USABLE_BITS_MASK: u64 = !(!0 << Self::TOTAL_USABLE_BITS);
    const SIGN_BITS_MASK: u64 = 0b111 << (Self::TOTAL_USABLE_BITS - 3);

    pub const fn from_raw(value: u64) -> Self {
        Self(value & Self::USABLE_BITS_MASK)
    }

    pub fn new(x: i32, y: i32, z: i32) -> Result<Self, ()> {
        if (x >= Self::ELEMENT_MIN && x <= Self::ELEMENT_MAX)
            && (y >= Self::ELEMENT_MIN && y <= Self::ELEMENT_MAX)
            && (z >= Self::ELEMENT_MIN && z <= Self::ELEMENT_MAX)
        {
            let raw = zorder::index_of([x as u32, y as u32, z as u32]);
            Ok(Self::from_raw(raw as u64))
        } else {
            Err(())
        }
    }

    pub fn raw(self) -> u64 {
        self.0
    }

    pub fn offset(self, x: i32, y: i32, z: i32) -> Result<Self, ()> {
        let (self_x, self_y, self_z) = self.into();
        Self::new(self_x + x, self_y + y, self_z + z)
    }
}

impl std::fmt::Display for ZOrderIndex {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (x, y, z) = (*self).into();
        write!(f, "[{}, {}, {}]", x, y, z)
    }
}

impl std::fmt::Debug for ZOrderIndex {
    fn fmt(&self, fmt: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (x, y, z) = (*self).into();
        fmt.debug_tuple(stringify!(ZOrderIndex))
            .field(&x)
            .field(&y)
            .field(&z)
            .finish()
    }
}

// Bitwise AND, OR, XOR

overload!((l: ZOrderIndex) & (r: u64) -> ZOrderIndex { ZOrderIndex::from_raw(l.0 & r) });
overload!((l: ZOrderIndex) | (r: u64) -> ZOrderIndex { ZOrderIndex::from_raw(l.0 | r) });
overload!((l: ZOrderIndex) ^ (r: u64) -> ZOrderIndex { ZOrderIndex::from_raw(l.0 ^ r) });
overload!((l: ZOrderIndex) & (r: ZOrderIndex) -> ZOrderIndex { l & r.0 });
overload!((l: ZOrderIndex) | (r: ZOrderIndex) -> ZOrderIndex { l | r.0 });
overload!((l: ZOrderIndex) ^ (r: ZOrderIndex) -> ZOrderIndex { l ^ r.0 });

// Bitshifting

overload!((l: ZOrderIndex) << (r: usize) -> ZOrderIndex {
    ZOrderIndex::from_raw(l.0 << (r * 3))
});

overload!((l: ZOrderIndex) >> (r: usize) -> ZOrderIndex {
    let mut result = l.0;
    let sign_bits = result & ZOrderIndex::SIGN_BITS_MASK;
    for _ in 0..r { result = (result >> 3) | sign_bits }
    ZOrderIndex::from_raw(result)
});

// Conversion

impl TryFrom<(i32, i32, i32)> for ZOrderIndex {
    type Error = ();
    fn try_from((x, y, z): (i32, i32, i32)) -> Result<Self, Self::Error> {
        Self::new(x, y, z)
    }
}
impl From<ZOrderIndex> for (i32, i32, i32) {
    fn from(value: ZOrderIndex) -> Self {
        const SHIFT: usize = (size_of::<i32>() * 8) - ZOrderIndex::BITS_PER_ELEMENT;
        zorder::coord_of(value.0 as u128)
            .map(|i| ((i as i32) << SHIFT) >> SHIFT)
            .into()
    }
}

impl TryFrom<[i32; 3]> for ZOrderIndex {
    type Error = ();
    fn try_from([x, y, z]: [i32; 3]) -> Result<Self, Self::Error> {
        Self::new(x, y, z)
    }
}
impl From<ZOrderIndex> for [i32; 3] {
    fn from(value: ZOrderIndex) -> Self {
        Into::<(i32, i32, i32)>::into(value).into()
    }
}

impl TryFrom<IVec3> for ZOrderIndex {
    type Error = ();
    fn try_from(IVec3 { x, y, z }: IVec3) -> Result<Self, Self::Error> {
        Self::new(x, y, z)
    }
}
impl From<ZOrderIndex> for IVec3 {
    fn from(value: ZOrderIndex) -> Self {
        Into::<(i32, i32, i32)>::into(value).into()
    }
}

impl TryFrom<ChunkPos> for ZOrderIndex {
    type Error = ();
    fn try_from(ChunkPos { x, y, z }: ChunkPos) -> Result<Self, Self::Error> {
        Self::new(x, y, z)
    }
}
impl From<ZOrderIndex> for ChunkPos {
    fn from(value: ZOrderIndex) -> Self {
        Into::<(i32, i32, i32)>::into(value).into()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn encode_decode_into() {
        assert_eq!((6, -16, 15), zorder(6, -16, 15).into());

        let min = ZOrderIndex::ELEMENT_MIN;
        let max = ZOrderIndex::ELEMENT_MAX;
        assert_eq!((min, 0, max), zorder(min, 0, max).into());
    }

    #[test]
    fn bounds_checking() {
        let min = 0b11111111111_1_00000000000000000000u32 as i32;
        let max = 0b00000000000_0_11111111111111111111u32 as i32;

        assert_eq!(
            Ok((ZOrderIndex::ELEMENT_MIN, ZOrderIndex::ELEMENT_MAX, 0)),
            ZOrderIndex::new(min, max, 0).map(Into::into)
        );

        assert!(ZOrderIndex::new(min - 1, max, 0).is_err());
        assert!(ZOrderIndex::new(min, max + 1, 0).is_err());
    }

    #[test]
    #[rustfmt::skip]
    fn raw() {
        let x_mask: u64 = 0b0_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_001_001_000;
        let y_mask: u64 = 0b0_010_010_010_010_010_010_010_010_010_010_010_010_010_010_010_010_010_000_000_000_000;
        let z_mask: u64 = 0b0_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_000_100_100_100_100;

        let raw = ZOrderIndex::from_raw(x_mask | y_mask | z_mask);
        assert_eq!(raw, zorder(6, -16, 15));
    }

    #[test]
    fn bitshift() {
        let zero = ZOrderIndex::ZERO;
        assert_eq!(zero, zero >> 2);
        assert_eq!(zero, zero << 2);

        assert_eq!(zorder(4, 8, 12), zorder(1, 2, 3) << 2);
        assert_eq!(zorder(1, 2, 3), zorder(4, 8, 12) >> 2);

        assert_eq!(zorder(-4, -8, -12), zorder(-1, -2, -3) << 2);
        assert_eq!(zorder(-1, -2, -3), zorder(-4, -8, -12) >> 2);
    }
}
