use bevy::{
    asset::RenderAssetUsages, prelude::*, render::mesh::Indices, render::mesh::PrimitiveTopology,
};

use crate::bloxel::{block::BlockTexture, prelude::*, storage::BloxelArray};

pub fn create_bloxel_mesh(
    bloxels: &impl BloxelView<Entity>,
    layout: &TextureAtlasLayout,
    sources: &TextureAtlasSources,
    block_lookup: &Query<&BlockTexture, With<Block>>,
) -> Mesh {
    let offset = -(bloxels.size().as_vec3() / 2.);

    let mut positions: Vec<Vec3> = vec![];
    let mut normals: Vec<Vec3> = vec![];
    let mut uvs: Vec<Vec2> = vec![];
    let mut indices = vec![];
    let mut num_vertices = 0;

    let mut append_vertex = |pos: Vec3, normal: Dir3, uv: Vec2| {
        positions.push(pos);
        normals.push(*normal);
        uvs.push(uv);
    };

    let mut append_quad = |show: bool, points: [Vec3; 4], normal: Dir3, uvs: Rect| {
        if show {
            append_vertex(points[0], normal, (uvs.min.x, uvs.min.y).into());
            append_vertex(points[1], normal, (uvs.max.x, uvs.min.y).into());
            append_vertex(points[2], normal, (uvs.max.x, uvs.max.y).into());
            append_vertex(points[3], normal, (uvs.min.x, uvs.max.y).into());

            for i in [0, 1, 2, 2, 3, 0] {
                indices.push(num_vertices + i);
            }
            num_vertices += 4;
        }
    };

    let mut append_cube = |pos: IVec3, uvs: Rect, sides: [bool; 6]| {
        let min = offset + pos.as_vec3();
        let max = offset + (pos + IVec3::ONE).as_vec3();

        // left/right, bottom/top, back/front
        let lbb = Vec3::new(min.x, min.y, min.z);
        let lbf = Vec3::new(min.x, max.y, max.z);
        let ltb = Vec3::new(min.x, max.y, min.z);
        let ltf = Vec3::new(min.x, min.y, max.z);
        let rbb = Vec3::new(max.x, min.y, min.z);
        let rbf = Vec3::new(max.x, max.y, max.z);
        let rtb = Vec3::new(max.x, max.y, min.z);
        let rtf = Vec3::new(max.x, min.y, max.z);

        append_quad(sides[0], [rbb, rtb, rbf, rtf], Dir3::X, uvs); // Right
        append_quad(sides[1], [ltf, lbf, ltb, lbb], Dir3::NEG_X, uvs); // Left
        append_quad(sides[2], [rtb, ltb, lbf, rbf], Dir3::Y, uvs); // Top
        append_quad(sides[3], [rtf, ltf, lbb, rbb], Dir3::NEG_Y, uvs); // Bottom
        append_quad(sides[4], [ltf, rtf, rbf, lbf], Dir3::Z, uvs); // Front
        append_quad(sides[5], [ltb, rtb, rbb, lbb], Dir3::NEG_Z, uvs); // Back
    };

    // Shrinks the UV rectangle by a tiny amount to get rid of ugly seams at the edges of blocks,
    // due to floating point inaccuracy when the color is looked up in the shader or something.
    // Temporary fix until a more sophisticated graphics overhaul using texture arrays or so.
    let shrink_amount = -0.1 / layout.size.x as f32;

    let size = bloxels.size().as_ivec3();

    // Capture all textures, which is currently all we need to know about a block to render it,
    // into a 3D array for us to look up, with a 1 block buffer so we can safely look up neighbors.
    let mut textures = BloxelArray::new_with_default(bloxels.size() + UVec3::new(2, 2, 2));
    for z in 0..size.z {
        for y in 0..size.y {
            for x in 0..size.x {
                let pos = IVec3::new(x, y, z);
                let block = bloxels.get(pos);
                textures[pos + IVec3::ONE] = block_lookup.get(block).ok().map(|t| &t.0);
            }
        }
    }

    for z in 0..size.z {
        for y in 0..size.y {
            for x in 0..size.x {
                let pos = IVec3::new(x, y, z);
                if let Some(texture) = textures[pos + IVec3::ONE] {
                    if let Some(uvs) = uv_rect(sources, layout, texture) {
                        append_cube(
                            pos,
                            uvs.inflate(shrink_amount),
                            [
                                textures[pos + IVec3::new(2, 1, 1)].is_none(),
                                textures[pos + IVec3::new(0, 1, 1)].is_none(),
                                textures[pos + IVec3::new(1, 2, 1)].is_none(),
                                textures[pos + IVec3::new(1, 0, 1)].is_none(),
                                textures[pos + IVec3::new(1, 1, 2)].is_none(),
                                textures[pos + IVec3::new(1, 1, 0)].is_none(),
                            ],
                        );
                    }
                }
            }
        }
    }

    Mesh::new(
        PrimitiveTopology::TriangleList,
        RenderAssetUsages::default(),
    )
    .with_inserted_attribute(Mesh::ATTRIBUTE_POSITION, positions)
    .with_inserted_attribute(Mesh::ATTRIBUTE_NORMAL, normals)
    .with_inserted_attribute(Mesh::ATTRIBUTE_UV_0, uvs)
    .with_inserted_indices(Indices::U32(indices))
}

fn uv_rect(
    sources: &TextureAtlasSources,
    layout: &TextureAtlasLayout,
    texture: impl Into<AssetId<Image>>,
) -> Option<Rect> {
    sources.texture_rect(layout, texture).map(|rect| {
        let rect = rect.as_rect();
        let size = layout.size.as_vec2();
        Rect::from_corners(rect.min / size, rect.max / size)
    })
}
