use bevy::prelude::*;

pub mod block;
pub mod chunk;
pub mod generic_math;
pub mod math;
pub mod mesh;
pub mod storage;
pub mod worldgen;

pub mod prelude {
    pub use super::{
        block::*,
        chunk::*,
        math::*,
        storage::{BloxelView, BloxelViewMut},
        worldgen::WorldGenPlugin,
        BloxelPlugin,
    };
}

use prelude::*;

pub struct BloxelPlugin;

impl Plugin for BloxelPlugin {
    fn build(&self, app: &mut App) {
        app.add_systems(
            Update,
            (set_transform_on_chunkpos_changed, mesh::generate_chunk_mesh),
        )
        .add_observer(add_chunks_to_chunkmap)
        .add_observer(remove_chunks_from_chunkmap);

        // Spawn default `ChunkMap` entity which all `Chunk`s get added to.
        app.world_mut().spawn(ChunkMap::default());
    }
}

fn set_transform_on_chunkpos_changed(
    mut commands: Commands,
    chunk_query: Query<(Entity, &ChunkPos), Changed<ChunkPos>>,
) {
    for (chunk, pos) in chunk_query.iter() {
        commands.entity(chunk).insert(pos.transform());
    }
}

fn add_chunks_to_chunkmap(
    trigger: Trigger<OnAdd, (Parent, ChunkPos)>,
    chunk_query: Query<(&Parent, &ChunkPos)>,
    mut map_query: Query<&mut ChunkMap>,
) {
    let chunk = trigger.entity();
    let Ok((parent, pos)) = chunk_query.get(chunk) else {
        // Trigger with a bundle doesn't work as expected. It triggers
        // when ANY component is added, rather than when ALL are present.
        return;
    };
    let mut map = map_query.get_mut(parent.get()).unwrap();
    map.try_insert(*pos, chunk).expect("chunk already present");
}

fn remove_chunks_from_chunkmap(
    trigger: Trigger<OnRemove, (Parent, ChunkPos)>,
    chunk_query: Query<(&Parent, &ChunkPos)>,
    mut map_query: Query<&mut ChunkMap>,
) {
    let chunk = trigger.entity();
    let Ok((parent, pos)) = chunk_query.get(chunk) else {
        // See above.
        return;
    };
    let mut map = map_query.get_mut(parent.get()).unwrap();
    map.remove(pos).expect("chunk not found");
    println!("Chunk {chunk} @ {pos:?} removed from {}", parent.get());
}
