use std::ops::{Index, IndexMut};

use bevy::math::{IVec3, UVec3};

use super::bloxel_view::{ivec3_to_index, BloxelView, BloxelViewMut};

pub struct BloxelArray<T: Copy> {
    size: UVec3,
    data: Vec<T>,
}

impl<T: Copy> BloxelArray<T> {
    pub fn new(size: UVec3, fill: T) -> Self {
        assert!(size.x > 0 && size.y > 0 && size.z > 0);
        let len = (size.x * size.y * size.z) as usize;
        let data = vec![fill; len];
        Self { size, data }
    }

    pub fn new_with_default(size: UVec3) -> Self
    where
        T: Default,
    {
        Self::new(size, Default::default())
    }

    fn get_index(&self, pos: impl Into<IVec3>) -> usize {
        let pos = pos.into();
        assert!(self.contains(pos));
        ivec3_to_index(pos, self.size)
    }
}

impl<T: Copy> Index<IVec3> for BloxelArray<T> {
    type Output = T;

    fn index(&self, pos: IVec3) -> &Self::Output {
        let index = self.get_index(pos);
        &self.data[index]
    }
}

impl<T: Copy> IndexMut<IVec3> for BloxelArray<T> {
    fn index_mut(&mut self, pos: IVec3) -> &mut Self::Output {
        let index = self.get_index(pos);
        &mut self.data[index]
    }
}

impl<T: Copy> BloxelView<T> for BloxelArray<T> {
    fn size(&self) -> UVec3 {
        self.size
    }

    fn contains(&self, pos: impl Into<IVec3>) -> bool {
        let pos = pos.into();
        let size = self.size.as_ivec3();
        (pos.x >= 0 && pos.x < size.x)
            && (pos.y >= 0 && pos.y < size.y)
            && (pos.z >= 0 && pos.z < size.z)
    }

    fn get(&self, pos: impl Into<IVec3>) -> T {
        let index = self.get_index(pos);
        self.data[index]
    }
}

impl<T: Copy> BloxelViewMut<T> for BloxelArray<T> {
    fn set(&mut self, pos: impl Into<IVec3>, value: T) {
        let index = self.get_index(pos);
        self.data[index] = value;
    }
}
