use std::{array::from_fn, collections::BinaryHeap};

use bevy::{
    math::IVec3,
    utils::{HashMap, HashSet},
};

use crate::bloxel::math::{zorder, ChunkPos, ChunkRegion, ZOrderIndex};

const START_INDEX_LOOKUP: [usize; 11] = [
    0, 1, 9, 73, 585, 4681, 37449, 299593, 2396745, 19173961, 153391689,
];

pub struct ChunkedOctree<T: Default + Copy + Eq> {
    depth: usize,
    regions: HashMap<ZOrderIndex, Vec<T>>,
}

impl<T: Default + Copy + Eq> ChunkedOctree<T> {
    pub fn new(depth: usize) -> Self {
        assert!(depth > 0 && depth < START_INDEX_LOOKUP.len());
        Self {
            depth,
            regions: Default::default(),
        }
    }

    pub fn get(&self, node: impl TryInto<OctreeNode>) -> T {
        if let Ok(node) = node.try_into() {
            if node.level <= self.depth {
                return self
                    .regions
                    .get(&(node.pos >> (self.depth - node.level)))
                    .map(|region| region[node.index(self.depth)])
                    .unwrap_or_default();
            }
        }
        Default::default()
    }

    pub fn update<F>(&mut self, chunk_pos: ChunkPos, update_fn: F)
    where
        F: Fn(OctreeNode, Option<&[T; 8]>, &mut T),
    {
        let pos: ZOrderIndex = chunk_pos.try_into().unwrap();
        let region = self.regions.entry(pos >> self.depth).or_insert_with(|| {
            let size = START_INDEX_LOOKUP[self.depth + 1] + 1;
            vec![Default::default(); size]
        });

        let mut node = OctreeNode::new(0, pos);
        while node.level <= self.depth {
            let parent_index = node.index(self.depth);
            let (parent, children) = if let Some(children_index) = node.children_index(self.depth) {
                let (left, right) = region.split_at_mut(children_index);
                let parent = &mut left[parent_index];
                let children = right[0..8].try_into().unwrap();
                (parent, Some(children))
            } else {
                let parent = &mut region[parent_index];
                (parent, None)
            };

            let previous = *parent;
            update_fn(node, children, parent);
            if *parent == previous {
                break; // If no change was made, we don't have anything to propagate.
            }

            node = node.parent();
        }
    }

    pub fn find<'a, W, F>(&'a self, priority_fn: F) -> OctreeIterator<'a, T, W, F>
    where
        W: Ord,
        F: Fn(OctreeNode, T) -> Option<W>,
    {
        OctreeIterator::new(self, priority_fn)
    }
}

pub struct OctreeIterator<'a, T, P, F>
where
    T: Default + Copy + Eq,
    P: Ord,
    F: Fn(OctreeNode, T) -> Option<P>,
{
    octree: &'a ChunkedOctree<T>,
    checked: HashSet<ZOrderIndex>,
    queue: BinaryHeap<PriorityItem<T, P>>,
    priority_fn: F,
}

impl<'a, T, P, F> OctreeIterator<'a, T, P, F>
where
    T: Default + Copy + Eq,
    P: Ord,
    F: Fn(OctreeNode, T) -> Option<P>,
{
    fn new(octree: &'a ChunkedOctree<T>, priority_fn: F) -> Self {
        let mut result = Self {
            octree,
            checked: Default::default(),
            queue: Default::default(),
            priority_fn,
        };
        for region_pos in octree.regions.keys() {
            result.search_region(*region_pos);
        }
        result
    }

    fn search_region(&mut self, region_pos: ZOrderIndex) {
        if self.checked.insert(region_pos) {
            self.push_item(OctreeNode::new(self.octree.depth, region_pos));
        }
    }

    fn push_item(&mut self, node: OctreeNode) {
        let value = self.octree.get(node);
        if let Some(priority) = (self.priority_fn)(node, value) {
            self.queue.push(PriorityItem {
                priority,
                node,
                value,
            });
        }
    }
}

impl<'a, T, P, F> Iterator for OctreeIterator<'a, T, P, F>
where
    T: Default + Copy + Eq,
    P: Ord,
    F: Fn(OctreeNode, T) -> Option<P>,
{
    type Item = (ChunkPos, T, P);

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            let PriorityItem {
                priority,
                node,
                value,
            } = self.queue.pop()?;

            if node.level == self.octree.depth {
                // If the current highest priority item is a region, see if its
                // neighboring regions are candidates to consider as well.
                // NOTE: Faulty `priority_fn` could cause this to loop forever.

                // Add all the region's neighbors to the priority queue.
                // NOTE: `search_region` skips any already added regions.
                for neighbor in node.neighbors() {
                    self.search_region(neighbor.pos);
                }
            }

            if let Some(children) = node.children() {
                // If current highest priority node has children,
                // add them to the priority queue and continue.
                for child in children {
                    self.push_item(child);
                }
            } else {
                // If there are no children, then we're at `ChunkPos` level.
                // Return this chunk as the highest priority item found.
                let chunk_pos = node.pos.into();
                return Some((chunk_pos, value, priority));
            }
        }
    }
}

#[derive(Default, Clone, Copy, PartialEq, Eq, Hash, Debug)]
pub struct OctreeNode {
    pub level: usize,
    pub pos: ZOrderIndex,
}

impl OctreeNode {
    pub fn new(level: usize, pos: ZOrderIndex) -> Self {
        Self { level, pos }
    }

    pub fn parent(&self) -> Self {
        Self::new(self.level + 1, self.pos >> 1)
    }

    pub fn children(&self) -> Option<[Self; 8]> {
        (self.level > 0).then(|| {
            let (level, pos) = (self.level - 1, self.pos << 1);
            from_fn(|i| Self::new(level, pos | i as u64))
        })
    }

    pub fn siblings(&self) -> [Self; 8] {
        let pos = self.pos & !0b111;
        from_fn(|i| Self::new(self.level, pos | i as u64))
    }

    pub fn neighbors(&self) -> [Self; 26] {
        #[rustfmt::skip]
        const DIRECTIONS: [(i32, i32, i32); 26] = [
            // Sides
            (-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1),
            // Edges
            (-1, -1,  0), (-1,  1,  0), (-1,  0, -1), (-1,  0,  1),
            ( 0, -1, -1), ( 0, -1,  1), ( 0,  1, -1), ( 0,  1,  1),
            ( 1, -1,  0), ( 1,  1,  0), ( 1,  0, -1), ( 1,  0,  1),
            // Corners
            (-1, -1, -1), ( 1, -1, -1), (-1,  1, -1), ( 1,  1, -1),
            (-1, -1,  1), ( 1, -1,  1), (-1,  1,  1), ( 1,  1,  1),
        ];

        let (pos_x, pos_y, pos_z) = self.pos.into();
        DIRECTIONS.map(|(x, y, z)| Self::new(self.level, zorder(pos_x + x, pos_y + y, pos_z + z)))
    }

    pub fn region(&self) -> ChunkRegion {
        let min = (self.pos << self.level).into();
        let max = min + IVec3::splat(!(!0 << self.level));
        ChunkRegion::new_unchecked(min, max)
    }

    fn index(&self, depth: usize) -> usize {
        let local_pos = self.pos & !(!0 << ((depth - self.level) * 3));
        START_INDEX_LOOKUP[depth - self.level] + local_pos.raw() as usize
    }

    fn children_index(&self, depth: usize) -> Option<usize> {
        (self.level > 0).then(|| Self::new(self.level - 1, self.pos << 1).index(depth))
    }
}

impl TryInto<OctreeNode> for ChunkPos {
    type Error = ();
    fn try_into(self) -> Result<OctreeNode, ()> {
        Ok(OctreeNode::new(0, self.try_into()?))
    }
}

impl TryInto<OctreeNode> for (i32, i32, i32) {
    type Error = ();
    fn try_into(self) -> Result<OctreeNode, ()> {
        Into::<ChunkPos>::into(self).try_into()
    }
}

struct PriorityItem<T: Default + Copy + Eq, P: Ord> {
    priority: P,
    node: OctreeNode,
    value: T,
}

impl<T: Default + Copy + Eq, P: Ord> Eq for PriorityItem<T, P> {}

impl<T: Default + Copy + Eq, P: Ord> PartialEq for PriorityItem<T, P> {
    fn eq(&self, other: &Self) -> bool {
        (self.priority == other.priority) && (self.node == other.node)
    }
}

impl<T: Default + Copy + Eq, P: Ord> Ord for PriorityItem<T, P> {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.priority.cmp(&other.priority)
    }
}

impl<T: Default + Copy + Eq, P: Ord> PartialOrd for PriorityItem<T, P> {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        self.priority.partial_cmp(&other.priority)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn update() {
        let node = |level: usize, pos: (i32, i32, i32)| -> OctreeNode {
            OctreeNode::new(level, pos.try_into().unwrap())
        };

        let mut octree = ChunkedOctree::<bool>::new(3);

        assert_eq!(false, octree.get((0, 0, 0)));
        assert_eq!(false, octree.get((1, 1, 1)));
        assert_eq!(false, octree.get((-1, -1, -1)));

        octree.update((0, 0, 0).into(), |_, _, parent| *parent = true);
        assert_eq!(true, octree.get((0, 0, 0)));

        assert_eq!(true, octree.get(node(0, (0, 0, 0))));
        assert_eq!(true, octree.get(node(1, (0, 0, 0))));
        assert_eq!(true, octree.get(node(2, (0, 0, 0))));
        assert_eq!(true, octree.get(node(3, (0, 0, 0))));

        assert_eq!(false, octree.get(node(0, (1, 1, 1))));
        assert_eq!(false, octree.get(node(1, (2, 2, 2))));
        assert_eq!(false, octree.get(node(2, (4, 4, 4))));
        assert_eq!(false, octree.get(node(3, (8, 8, 8))));

        assert_eq!(false, octree.get(node(0, (-1, -1, -1))));
        assert_eq!(false, octree.get(node(1, (-1, -1, -1))));
        assert_eq!(false, octree.get(node(2, (-1, -1, -1))));
        assert_eq!(false, octree.get(node(3, (-1, -1, -1))));

        octree.update(ChunkPos::new(-12, -17, -42), |_, _, parent| *parent = true);
        assert_eq!(true, octree.get((-12, -17, -42)));

        assert_eq!(true, octree.get(node(0, (-12, -17, -42))));
        assert_eq!(true, octree.get(node(1, (-6, -9, -21))));
        assert_eq!(true, octree.get(node(2, (-3, -5, -11))));
        assert_eq!(true, octree.get(node(3, (-2, -3, -6))));
    }

    #[test]
    #[rustfmt::skip]
    fn find() {
        let mut octree = ChunkedOctree::<bool>::new(3);
        octree.update(( 2,  3,  4).into(), |_, _, parent| *parent = true);
        octree.update((10, 10, 10).into(), |_, _, parent| *parent = true);
        octree.update(( 8,  8,  8).into(), |_, _, parent| *parent = true);
        octree.update(( 0, 16, -1).into(), |_, _, parent| *parent = true);
        octree.update(( 9,  9,  9).into(), |_, _, parent| *parent = true);

        let pos = ChunkPos::new(8, 8, 8);
        let mut iterator = octree.find(|node, value| {
            value.then(|| -node.region().distance_to_squared(pos))
        });

        assert_eq!(Some((( 8,  8,  8).into(), true, -(0*0 + 0*0 + 0*0))), iterator.next());
        assert_eq!(Some((( 9,  9,  9).into(), true, -(1*1 + 1*1 + 1*1))), iterator.next());
        assert_eq!(Some(((10, 10, 10).into(), true, -(2*2 + 2*2 + 2*2))), iterator.next());
        assert_eq!(Some((( 2,  3,  4).into(), true, -(6*6 + 5*5 + 4*4))), iterator.next());
        assert_eq!(Some((( 0, 16, -1).into(), true, -(8*8 + 8*8 + 9*9))), iterator.next());
        assert_eq!(None, iterator.next());
    }

    #[test]
    fn node_region() {
        let chunk = |a: (i32, i32, i32), b: (i32, i32, i32)| -> ChunkRegion {
            ChunkRegion::new(a.into(), b.into())
        };
        let node = |level: usize, pos: (i32, i32, i32)| -> ChunkRegion {
            OctreeNode::new(level, pos.try_into().unwrap()).region()
        };

        assert_eq!(chunk((0, 0, 0), (0, 0, 0)), node(0, (0, 0, 0)));
        assert_eq!(chunk((0, 0, 0), (1, 1, 1)), node(1, (0, 0, 0)));
        assert_eq!(chunk((0, 0, 0), (3, 3, 3)), node(2, (0, 0, 0)));
        assert_eq!(chunk((0, 0, 0), (7, 7, 7)), node(3, (0, 0, 0)));

        assert_eq!(chunk((-1, -1, -1), (-1, -1, -1)), node(0, (-1, -1, -1)));
        assert_eq!(chunk((-2, -2, -2), (-1, -1, -1)), node(1, (-1, -1, -1)));
        assert_eq!(chunk((-4, -4, -4), (-1, -1, -1)), node(2, (-1, -1, -1)));
        assert_eq!(chunk((-8, -8, -8), (-1, -1, -1)), node(3, (-1, -1, -1)));

        assert_eq!(chunk((2, -3, -4), (2, -3, -4)), node(0, (2, -3, -4)));
        assert_eq!(chunk((4, -6, -8), (5, -5, -7)), node(1, (2, -3, -4)));
        assert_eq!(chunk((8, -12, -16), (11, -9, -13)), node(2, (2, -3, -4)));
    }
}
