use bevy::math::{IVec3, UVec3};

use super::{
    bloxel_view::{ivec3_to_index, BloxelView, BloxelViewMut},
    palette_storage::PaletteStorage,
};

pub struct PaletteBloxelStorage<T: Copy + Eq> {
    size: UVec3,
    data: PaletteStorage<T>,
}

impl<T: Copy + Eq> PaletteBloxelStorage<T> {
    pub fn new(size: UVec3, fill: T) -> Self {
        assert!(size.x > 0 && size.y > 0 && size.z > 0);
        let len = (size.x * size.y * size.z).try_into().unwrap();
        let data = PaletteStorage::new(len, fill);
        Self { size, data }
    }

    pub fn new_with_default(size: UVec3) -> Self
    where
        T: Default,
    {
        Self::new(size, Default::default())
    }

    pub fn get_used_count(&self, value: T) -> usize {
        self.data.get_used_count(value)
    }

    fn get_index(&self, pos: impl Into<IVec3>) -> usize {
        let pos = pos.into();
        assert!(self.contains(pos));
        ivec3_to_index(pos, self.size)
    }
}

impl<T: Copy + Eq> BloxelView<T> for PaletteBloxelStorage<T> {
    fn size(&self) -> UVec3 {
        self.size
    }

    fn contains(&self, pos: impl Into<IVec3>) -> bool {
        let pos = pos.into();
        let size = self.size.as_ivec3();
        (pos.x >= 0 && pos.x < size.x)
            && (pos.y >= 0 && pos.y < size.y)
            && (pos.z >= 0 && pos.z < size.z)
    }

    fn get(&self, pos: impl Into<IVec3>) -> T {
        let index = self.get_index(pos);
        self.data.get(index)
    }
}

impl<T: Copy + Eq> BloxelViewMut<T> for PaletteBloxelStorage<T> {
    fn set(&mut self, pos: impl Into<IVec3>, value: T) {
        let index = self.get_index(pos);
        self.data.set(index, value);
    }
}
