use bevy::{asset::LoadedFolder, prelude::*};

mod bloxel;
mod camera_controller;

use bloxel::block::BlockTexture;
use bloxel::prelude::*;
use camera_controller::{CameraControllerPlugin, ControlledCamera};

#[derive(Resource)]
pub struct TerrainTextures(Handle<LoadedFolder>);

#[derive(Resource)]
pub struct TerrainAtlas {
    pub layout: Handle<TextureAtlasLayout>,
    pub sources: TextureAtlasSources,
    pub texture: Handle<Image>,
}

#[derive(Resource)]
pub struct TerrainMaterial(Handle<StandardMaterial>);

#[derive(Resource)]
pub struct TerrainBlocks {
    air: Entity,
    grass: Entity,
    dirt: Entity,
    rock: Entity,
    sand: Entity,
}

fn main() {
    App::new()
        .add_plugins((
            DefaultPlugins.set(ImagePlugin::default_nearest()),
            CameraControllerPlugin,
            BloxelPlugin,
            WorldGenPlugin,
        ))
        .add_systems(
            Startup,
            (
                setup_sunlight,
                setup_camera,
                setup_terrain_blocks,
                load_terrain_textures,
            ),
        )
        .add_systems(Update, check_terrain_textures)
        .run();
}

fn setup_terrain_blocks(mut commands: Commands, asset_server: Res<AssetServer>) {
    let air = commands.spawn((Block, Name::new("air"))).id();

    let mut block = |name: &'static str, texture: &str| -> Entity {
        let path = format!("terrain/{texture}.png");
        let texture = BlockTexture(asset_server.load(path));
        commands.spawn((Block, Name::new(name), texture)).id()
    };

    let blocks = TerrainBlocks {
        air,
        grass: block("grass", "grass_top"),
        dirt: block("dirt", "dirt"),
        rock: block("rock", "rock"),
        sand: block("sand", "sand"),
    };

    commands.insert_resource(blocks);
}

fn load_terrain_textures(mut commands: Commands, asset_server: Res<AssetServer>) {
    commands.insert_resource(TerrainTextures(asset_server.load_folder("terrain")));
}

/// Waits for all textures in `assets/terrain/` to be loaded, then builds an atlas
/// from those along with the [`TerrainMaterial`] and [`TerrainAtlas`] resources.
fn check_terrain_textures(
    mut commands: Commands,
    mut events: EventReader<AssetEvent<LoadedFolder>>,
    terrain_textures: Res<TerrainTextures>,
    loaded_folders: Res<Assets<LoadedFolder>>,
    mut textures: ResMut<Assets<Image>>,
    mut materials: ResMut<Assets<StandardMaterial>>,
    mut atlas_layouts: ResMut<Assets<TextureAtlasLayout>>,
) {
    for event in events.read() {
        if event.is_loaded_with_dependencies(&terrain_textures.0) {
            let mut atlas_builder = TextureAtlasBuilder::default();

            let folder = loaded_folders.get(&terrain_textures.0).unwrap();
            for handle in folder.handles.iter() {
                let id = handle.id().typed_unchecked::<Image>();
                let Some(texture) = textures.get(id) else {
                    warn!(
                        "{:?} did not resolve to an `Image` asset.",
                        handle.path().unwrap()
                    );
                    continue;
                };
                atlas_builder.add_texture(Some(id), texture);
            }

            let (layout, sources, texture) = atlas_builder.build().unwrap();
            let layout = atlas_layouts.add(layout);
            let texture = textures.add(texture);
            let material = materials.add(texture.clone());

            commands.insert_resource(TerrainMaterial(material));

            commands.insert_resource(TerrainAtlas {
                layout,
                sources,
                texture,
            });
        }
    }
}

fn setup_sunlight(mut commands: Commands) {
    commands.spawn((
        DirectionalLight {
            shadows_enabled: true,
            ..default()
        },
        Transform::from_xyz(4., 8., 4.).looking_at(Vec3::ZERO, Vec3::Y),
    ));
}

fn setup_camera(mut commands: Commands) {
    commands.spawn((
        ControlledCamera,
        Camera3d::default(),
        Transform::from_xyz(-24., 16., 16.).looking_at((0., 0., 0.).into(), Vec3::Y),
    ));
}
